@@ -19,7 +19,6 @@
 */
 
 #include "includes.h"
-#include "lib/events/events.h"
 #include "system/filesys.h"
 #include "system/wait.h"
 #include "db_wrap.h"
@@ -28,6 +27,8 @@
 
 struct ctdb_persistent_state {
 	struct ctdb_context *ctdb;
+	struct ctdb_db_context *ctdb_db; /* used by trans3_commit */
+	struct ctdb_client *client; /* used by trans3_commit */
 	struct ctdb_req_control *c;
 	const char *errormsg;
 	uint32_t num_pending;
@@ -52,27 +53,48 @@ static void ctdb_persistent_callback(struct ctdb_context *ctdb,
 {
 	struct ctdb_persistent_state *state = talloc_get_type(private_data, 
 							      struct ctdb_persistent_state);
+	enum ctdb_trans2_commit_error etype;
+
+	if (ctdb->recovery_mode != CTDB_RECOVERY_NORMAL) {
+		DEBUG(DEBUG_INFO, ("ctdb_persistent_callback: ignoring reply "
+				   "during recovery\n"));
+		return;
+	}
 
 	if (status != 0) {
 		DEBUG(DEBUG_ERR,("ctdb_persistent_callback failed with status %d (%s)\n",
-			 status, errormsg));
+			 status, errormsg?errormsg:"no error message given"));
 		state->status = status;
 		state->errormsg = errormsg;
 		state->num_failed++;
+
+		/*
+		 * If a node failed to complete the update_record control,
+		 * then either a recovery is already running or something
+		 * bad is going on. So trigger a recovery and let the
+		 * recovery finish the transaction, sending back the reply
+		 * for the trans3_commit control to the client.
+		 */
+		ctdb->recovery_mode = CTDB_RECOVERY_ACTIVE;
+		return;
 	}
+
 	state->num_pending--;
-	if (state->num_pending == 0) {
-		enum ctdb_trans2_commit_error etype;
-		if (state->num_failed == state->num_sent) {
-			etype = CTDB_TRANS2_COMMIT_ALLFAIL;
-		} else if (state->num_failed != 0) {
-			etype = CTDB_TRANS2_COMMIT_SOMEFAIL;
-		} else {
-			etype = CTDB_TRANS2_COMMIT_SUCCESS;
-		}
-		ctdb_request_control_reply(state->ctdb, state->c, NULL, etype, state->errormsg);
-		talloc_free(state);
+
+	if (state->num_pending != 0) {
+		return;
 	}
+
+	if (state->num_failed == state->num_sent) {
+		etype = CTDB_TRANS2_COMMIT_ALLFAIL;
+	} else if (state->num_failed != 0) {
+		etype = CTDB_TRANS2_COMMIT_SOMEFAIL;
+	} else {
+		etype = CTDB_TRANS2_COMMIT_SUCCESS;
+	}
+
+	ctdb_request_control_reply(state->ctdb, state->c, NULL, etype, state->errormsg);
+	talloc_free(state);
 }
 
 /*
@@ -82,13 +104,53 @@ static void ctdb_persistent_store_timeout(struct event_context *ev, struct timed
 					 struct timeval t, void *private_data)
 {
 	struct ctdb_persistent_state *state = talloc_get_type(private_data, struct ctdb_persistent_state);
-	
+
+	if (state->ctdb->recovery_mode != CTDB_RECOVERY_NORMAL) {
+		DEBUG(DEBUG_INFO, ("ctdb_persistent_store_timeout: ignoring "
+				   "timeout during recovery\n"));
+		return;
+	}
+
 	ctdb_request_control_reply(state->ctdb, state->c, NULL, CTDB_TRANS2_COMMIT_TIMEOUT, 
 				   "timeout in ctdb_persistent_state");
 
 	talloc_free(state);
 }
 
+/**
+ * Finish pending trans3 commit controls, i.e. send
+ * reply to the client. This is called by the end-recovery
+ * control to fix the situation when a recovery interrupts
+ * the usual porgress of a transaction.
+ */
+void ctdb_persistent_finish_trans3_commits(struct ctdb_context *ctdb)
+{
+	struct ctdb_db_context *ctdb_db;
+
+	if (ctdb->recovery_mode != CTDB_RECOVERY_NORMAL) {
+		DEBUG(DEBUG_INFO, ("ctdb_persistent_store_timeout: ignoring "
+				   "timeout during recovery\n"));
+		return;
+	}
+
+	for (ctdb_db = ctdb->db_list; ctdb_db; ctdb_db = ctdb_db->next) {
+		struct ctdb_persistent_state *state;
+
+		if (ctdb_db->persistent_state == NULL) {
+			continue;
+		}
+
+		state = ctdb_db->persistent_state;
+
+		ctdb_request_control_reply(ctdb, state->c, NULL,
+					   CTDB_TRANS2_COMMIT_SOMEFAIL,
+					   "trans3 commit ended by recovery");
+
+		/* The destructor sets ctdb_db->persistent_state to NULL. */
+		talloc_free(state);
+	}
+}
+
 /*
   store a set of persistent records - called from a ctdb client when it has updated
   some records in a persistent database. The client will have the record
@@ -117,6 +179,12 @@ int32_t ctdb_control_trans2_commit(struct ctdb_context *ctdb,
 		return -1;
 	}
 
+	if (ctdb_db->unhealthy_reason) {
+		DEBUG(DEBUG_ERR,("db(%s) unhealty in ctdb_control_trans2_commit: %s\n",
+				 ctdb_db->db_name, ctdb_db->unhealthy_reason));
+		return -1;
+	}
+
 	/* handling num_persistent_updates is a bit strange - 
 	   there are 3 cases
 	     1) very old clients, which never called CTDB_CONTROL_START_PERSISTENT_UPDATE
@@ -241,294 +309,113 @@ int32_t ctdb_control_trans2_commit(struct ctdb_context *ctdb,
 	return 0;
 }
 
-
-struct ctdb_persistent_write_state {
-	struct ctdb_db_context *ctdb_db;
-	struct ctdb_marshall_buffer *m;
-	struct ctdb_req_control *c;
-};
-
-
-/*
-  called from a child process to write the data
- */
-static int ctdb_persistent_store(struct ctdb_persistent_write_state *state)
+static int ctdb_persistent_state_destructor(struct ctdb_persistent_state *state)
 {
-	int ret, i;
-	struct ctdb_rec_data *rec = NULL;
-	struct ctdb_marshall_buffer *m = state->m;
-
-	ret = tdb_transaction_start(state->ctdb_db->ltdb->tdb);
-	if (ret == -1) {
-		DEBUG(DEBUG_ERR,("Failed to start transaction for db_id 0x%08x in ctdb_persistent_store\n",
-				 state->ctdb_db->db_id));
-		return -1;
+	if (state->client != NULL) {
+		state->client->db_id = 0;
 	}
 
-	for (i=0;i<m->count;i++) {
-		struct ctdb_ltdb_header oldheader;
-		struct ctdb_ltdb_header header;
-		TDB_DATA key, data, olddata;
-		TALLOC_CTX *tmp_ctx = talloc_new(state);
-
-		rec = ctdb_marshall_loop_next(m, rec, NULL, &header, &key, &data);
-		
-		if (rec == NULL) {
-			DEBUG(DEBUG_ERR,("Failed to get next record %d for db_id 0x%08x in ctdb_persistent_store\n",
-					 i, state->ctdb_db->db_id));
-			talloc_free(tmp_ctx);
-			goto failed;			
-		}
-
-		/* fetch the old header and ensure the rsn is less than the new rsn */
-		ret = ctdb_ltdb_fetch(state->ctdb_db, key, &oldheader, tmp_ctx, &olddata);
-		if (ret != 0) {
-			DEBUG(DEBUG_ERR,("Failed to fetch old record for db_id 0x%08x in ctdb_persistent_store\n",
-					 state->ctdb_db->db_id));
-			talloc_free(tmp_ctx);
-			goto failed;
-		}
-
-		if (oldheader.rsn >= header.rsn &&
-		    (olddata.dsize != data.dsize || 
-		     memcmp(olddata.dptr, data.dptr, data.dsize) != 0)) {
-			DEBUG(DEBUG_CRIT,("existing header for db_id 0x%08x has larger RSN %llu than new RSN %llu in ctdb_persistent_store\n",
-					  state->ctdb_db->db_id, 
-					  (unsigned long long)oldheader.rsn, (unsigned long long)header.rsn));
-			talloc_free(tmp_ctx);
-			goto failed;
-		}
-
-		talloc_free(tmp_ctx);
-
-		ret = ctdb_ltdb_store(state->ctdb_db, key, &header, data);
-		if (ret != 0) {
-			DEBUG(DEBUG_CRIT,("Failed to store record for db_id 0x%08x in ctdb_persistent_store\n", 
-					  state->ctdb_db->db_id));
-			goto failed;
-		}
-	}
-
-	ret = tdb_transaction_commit(state->ctdb_db->ltdb->tdb);
-	if (ret == -1) {
-		DEBUG(DEBUG_ERR,("Failed to commit transaction for db_id 0x%08x in ctdb_persistent_store\n",
-				 state->ctdb_db->db_id));
-		return -1;
+	if (state->ctdb_db != NULL) {
+		state->ctdb_db->persistent_state = NULL;
 	}
 
 	return 0;
-	
-failed:
-	tdb_transaction_cancel(state->ctdb_db->ltdb->tdb);
-	return -1;
-}
-
-
-/*
-  called when we the child has completed the persistent write
-  on our behalf
- */
-static void ctdb_persistent_write_callback(int status, void *private_data)
-{
-	struct ctdb_persistent_write_state *state = talloc_get_type(private_data, 
-								   struct ctdb_persistent_write_state);
-
-
-	ctdb_request_control_reply(state->ctdb_db->ctdb, state->c, NULL, status, NULL);
-
-	talloc_free(state);
 }
 
 /*
-  called if our lockwait child times out
+ * Store a set of persistent records.
+ * This is used to roll out a transaction to all nodes.
  */
-static void ctdb_persistent_lock_timeout(struct event_context *ev, struct timed_event *te, 
-					 struct timeval t, void *private_data)
+int32_t ctdb_control_trans3_commit(struct ctdb_context *ctdb,
+				   struct ctdb_req_control *c,
+				   TDB_DATA recdata, bool *async_reply)
 {
-	struct ctdb_persistent_write_state *state = talloc_get_type(private_data, 
-								   struct ctdb_persistent_write_state);
-	ctdb_request_control_reply(state->ctdb_db->ctdb, state->c, NULL, -1, "timeout in ctdb_persistent_lock");
-	talloc_free(state);
-}
-
-struct childwrite_handle {
-	struct ctdb_context *ctdb;
-	struct ctdb_db_context *ctdb_db;
-	struct fd_event *fde;
-	int fd[2];
-	pid_t child;
-	void *private_data;
-	void (*callback)(int, void *);
-	struct timeval start_time;
-};
-
-static int childwrite_destructor(struct childwrite_handle *h)
-{
-	h->ctdb->statistics.pending_childwrite_calls--;
-	kill(h->child, SIGKILL);
-	return 0;
-}
-
-/* called when the child process has finished writing the record to the
-   database
-*/
-static void childwrite_handler(struct event_context *ev, struct fd_event *fde, 
-			     uint16_t flags, void *private_data)
-{
-	struct childwrite_handle *h = talloc_get_type(private_data, 
-						     struct childwrite_handle);
-	void *p = h->private_data;
-	void (*callback)(int, void *) = h->callback;
-	pid_t child = h->child;
-	TALLOC_CTX *tmp_ctx = talloc_new(ev);
-	int ret;
-	char c;
-
-	ctdb_latency(h->ctdb_db, "persistent", &h->ctdb->statistics.max_childwrite_latency, h->start_time);
-	h->ctdb->statistics.pending_childwrite_calls--;
-
-	/* the handle needs to go away when the context is gone - when
-	   the handle goes away this implicitly closes the pipe, which
-	   kills the child */
-	talloc_steal(tmp_ctx, h);
-
-	talloc_set_destructor(h, NULL);
-
-	ret = read(h->fd[0], &c, 1);
-	if (ret < 1) {
-		DEBUG(DEBUG_ERR, (__location__ " Read returned %d. Childwrite failed\n", ret));
-		c = 1;
-	}
-
-	callback(c, p);
-
-	kill(child, SIGKILL);
-	talloc_free(tmp_ctx);
-}
-
-/* this creates a child process which will take out a tdb transaction
-   and write the record to the database.
-*/
-struct childwrite_handle *ctdb_childwrite(struct ctdb_db_context *ctdb_db,
-				void (*callback)(int, void *private_data),
-				struct ctdb_persistent_write_state *state)
-{
-	struct childwrite_handle *result;
-	int ret;
-	pid_t parent = getpid();
-
-	ctdb_db->ctdb->statistics.childwrite_calls++;
-	ctdb_db->ctdb->statistics.pending_childwrite_calls++;
-
-	if (!(result = talloc_zero(state, struct childwrite_handle))) {
-		ctdb_db->ctdb->statistics.pending_childwrite_calls--;
-		return NULL;
-	}
-
-	ret = pipe(result->fd);
-
-	if (ret != 0) {
-		talloc_free(result);
-		ctdb_db->ctdb->statistics.pending_childwrite_calls--;
-		return NULL;
-	}
-
-	result->child = fork();
-
-	if (result->child == (pid_t)-1) {
-		close(result->fd[0]);
-		close(result->fd[1]);
-		talloc_free(result);
-		ctdb_db->ctdb->statistics.pending_childwrite_calls--;
-		return NULL;
-	}
-
-	result->callback = callback;
-	result->private_data = state;
-	result->ctdb = ctdb_db->ctdb;
-	result->ctdb_db = ctdb_db;
-
-	if (result->child == 0) {
-		char c = 0;
-
-		close(result->fd[0]);
-		ret = ctdb_persistent_store(state);
-		if (ret != 0) {
-			DEBUG(DEBUG_ERR, (__location__ " Failed to write persistent data\n"));
-			c = 1;
-		}
-
-		write(result->fd[1], &c, 1);
-
-		/* make sure we die when our parent dies */
-		while (kill(parent, 0) == 0 || errno != ESRCH) {
-			sleep(5);
-		}
-		_exit(0);
-	}
-
-	close(result->fd[1]);
-	set_close_on_exec(result->fd[0]);
-
-	talloc_set_destructor(result, childwrite_destructor);
-
-	DEBUG(DEBUG_NOTICE, (__location__ " Created PIPE FD:%d for ctdb_childwrite\n", result->fd[0]));
-
-	result->fde = event_add_fd(ctdb_db->ctdb->ev, result, result->fd[0],
-				   EVENT_FD_READ|EVENT_FD_AUTOCLOSE, childwrite_handler,
-				   (void *)result);
-	if (result->fde == NULL) {
-		talloc_free(result);
-		ctdb_db->ctdb->statistics.pending_childwrite_calls--;
-		return NULL;
-	}
-
-	result->start_time = timeval_current();
-
-	return result;
-}
-
-/* 
-   update a record on this node if the new record has a higher rsn than the
-   current record
- */
-int32_t ctdb_control_update_record(struct ctdb_context *ctdb, 
-				   struct ctdb_req_control *c, TDB_DATA recdata, 
-				   bool *async_reply)
-{
-	struct ctdb_db_context *ctdb_db;
-	struct ctdb_persistent_write_state *state;
-	struct childwrite_handle *handle;
+	struct ctdb_client *client;
+	struct ctdb_persistent_state *state;
+	int i;
 	struct ctdb_marshall_buffer *m = (struct ctdb_marshall_buffer *)recdata.dptr;
+	struct ctdb_db_context *ctdb_db;
 
 	if (ctdb->recovery_mode != CTDB_RECOVERY_NORMAL) {
-		DEBUG(DEBUG_INFO,("rejecting ctdb_control_update_record when recovery active\n"));
+		DEBUG(DEBUG_INFO,("rejecting ctdb_control_trans3_commit when recovery active\n"));
+		return -1;
+	}
+
+	client = ctdb_reqid_find(ctdb, c->client_id, struct ctdb_client);
+	if (client == NULL) {
+		DEBUG(DEBUG_ERR,(__location__ " can not match persistent_store "
+				 "to a client. Returning error\n"));
+		return -1;
+	}
+
+	if (client->db_id != 0) {
+		DEBUG(DEBUG_ERR,(__location__ " ERROR: trans3_commit: "
+				 "client-db_id[0x%08x] != 0 "
+				 "(client_id[0x%08x]): trans3_commit active?\n",
+				 client->db_id, client->client_id));
 		return -1;
 	}
 
 	ctdb_db = find_ctdb_db(ctdb, m->db_id);
 	if (ctdb_db == NULL) {
-		DEBUG(DEBUG_ERR,("Unknown database 0x%08x in ctdb_control_update_record\n", m->db_id));
+		DEBUG(DEBUG_ERR,(__location__ " ctdb_control_trans3_commit: "
+				 "Unknown database db_id[0x%08x]\n", m->db_id));
 		return -1;
 	}
 
-	state = talloc(ctdb, struct ctdb_persistent_write_state);
-	CTDB_NO_MEMORY(ctdb, state);
-
-	state->ctdb_db = ctdb_db;
-	state->c       = c;
-	state->m       = m;
-
-	/* create a child process to take out a transaction and 
-	   write the data.
-	*/
-	handle = ctdb_childwrite(ctdb_db, ctdb_persistent_write_callback, state);
-	if (handle == NULL) {
-		DEBUG(DEBUG_ERR,("Failed to setup childwrite handler in ctdb_control_update_record\n"));
-		talloc_free(state);
+	if (ctdb_db->persistent_state != NULL) {
+		DEBUG(DEBUG_ERR, (__location__ " Error: "
+				  "ctdb_control_trans3_commit "
+				  "called while a transaction commit is "
+				  "active. db_id[0x%08x]\n", m->db_id));
 		return -1;
 	}
 
+	ctdb_db->persistent_state = talloc_zero(ctdb_db,
+						struct ctdb_persistent_state);
+	CTDB_NO_MEMORY(ctdb, ctdb_db->persistent_state);
+
+	client->db_id = m->db_id;
+
+	state = ctdb_db->persistent_state;
+	state->ctdb = ctdb;
+	state->ctdb_db = ctdb_db;
+	state->c    = c;
+	state->client = client;
+
+	talloc_set_destructor(state, ctdb_persistent_state_destructor);
+
+	for (i = 0; i < ctdb->vnn_map->size; i++) {
+		struct ctdb_node *node = ctdb->nodes[ctdb->vnn_map->map[i]];
+		int ret;
+
+		/* only send to active nodes */
+		if (node->flags & NODE_FLAGS_INACTIVE) {
+			continue;
+		}
+
+		ret = ctdb_daemon_send_control(ctdb, node->pnn, 0,
+					       CTDB_CONTROL_UPDATE_RECORD,
+					       c->client_id, 0, recdata,
+					       ctdb_persistent_callback,
+					       state);
+		if (ret == -1) {
+			DEBUG(DEBUG_ERR,("Unable to send "
+					 "CTDB_CONTROL_UPDATE_RECORD "
+					 "to pnn %u\n", node->pnn));
+			talloc_free(state);
+			return -1;
+		}
+
+		state->num_pending++;
+		state->num_sent++;
+	}
+
+	if (state->num_pending == 0) {
+		talloc_free(state);
+		return 0;
+	}
+
 	/* we need to wait for the replies */
 	*async_reply = true;
 
@@ -536,8 +423,9 @@ int32_t ctdb_control_update_record(struct ctdb_context *ctdb,
 	talloc_steal(state, c);
 
 	/* but we won't wait forever */
-	event_add_timed(ctdb->ev, state, timeval_current_ofs(ctdb->tunable.control_timeout, 0),
-			ctdb_persistent_lock_timeout, state);
+	event_add_timed(ctdb->ev, state,
+			timeval_current_ofs(ctdb->tunable.control_timeout, 0),
+			ctdb_persistent_store_timeout, state);
 
 	return 0;
 }
@@ -730,4 +618,71 @@ int32_t ctdb_control_persistent_store(struct ctdb_context *ctdb,
 	return ctdb_control_trans2_commit(ctdb, c, ctdb_marshall_finish(m), async_reply);
 }
 
+static int32_t ctdb_get_db_seqnum(struct ctdb_context *ctdb,
+				  uint32_t db_id,
+				  uint64_t *seqnum)
+{
+	int32_t ret;
+	struct ctdb_db_context *ctdb_db;
+	const char *keyname = CTDB_DB_SEQNUM_KEY;
+	TDB_DATA key;
+	TDB_DATA data;
+	TALLOC_CTX *mem_ctx = talloc_new(ctdb);
+	struct ctdb_ltdb_header header;
 
+	ctdb_db = find_ctdb_db(ctdb, db_id);
+	if (!ctdb_db) {
+		DEBUG(DEBUG_ERR,(__location__ " Unknown db 0x%08x\n", db_id));
+		ret = -1;
+		goto done;
+	}
+
+	key.dptr = (uint8_t *)discard_const(keyname);
+	key.dsize = strlen(keyname) + 1;
+
+	ret = (int32_t)ctdb_ltdb_fetch(ctdb_db, key, &header, mem_ctx, &data);
+	if (ret != 0) {
+		goto done;
+	}
+
+	if (data.dsize != sizeof(uint64_t)) {
+		*seqnum = 0;
+		goto done;
+	}
+
+	*seqnum = *(uint64_t *)data.dptr;
+
+done:
+	talloc_free(mem_ctx);
+	return ret;
+}
+
+/**
+ * Get the sequence number of a persistent database.
+ */
+int32_t ctdb_control_get_db_seqnum(struct ctdb_context *ctdb,
+				   TDB_DATA indata,
+				   TDB_DATA *outdata)
+{
+	uint32_t db_id;
+	int32_t ret;
+	uint64_t seqnum;
+
+	db_id = *(uint32_t *)indata.dptr;
+	ret = ctdb_get_db_seqnum(ctdb, db_id, &seqnum);
+	if (ret != 0) {
+		goto done;
+	}
+
+	outdata->dsize = sizeof(uint64_t);
+	outdata->dptr = (uint8_t *)talloc_zero(outdata, uint64_t);
+	if (outdata->dptr == NULL) {
+		ret = -1;
+		goto done;
+	}
+
+	*(outdata->dptr) = seqnum;
+
+done:
+	return ret;
+}
